/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal GUI (http://www.irisa.fr/espresso/polychrony).

   Polychrony Signal GUI is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal GUI is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal GUI; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
 */
//_PUBLIC_


/**
  Constructor
*/
graphicTablesManager::graphicTablesManager ()
{
  int i;

  for (i = 0; i < taille_tab; i++)
    t_acces_seg[i] = null;
  for (i = 0; i < taille_tab; i++)
    tab_acces[i] = null;
  the_freeGraphObjIdentity = 1;
}


/**
  Allocate a new (unique) number for boxes, ports, paths, links or segments
*/
graphObj_type
graphicTablesManager::newGraphObjIdentity ()
{
  graphObj_type n;
  if (the_freeGraphObjIdentity >= GRAPHOBJ_TYPEMAX)
    {
      fprintf (stderr, "###INTERNAL ERROR###: too large graph object identification\n");
      fflush (stderr);
      return (graphObj_type) - 1;
    }
  n = the_freeGraphObjIdentity;
  the_freeGraphObjIdentity++;
  return (graphObj_type) n;
}


/**
  Returns the data structure associated with the object obj (in).
  obj may be a box, a port, a path or a link, but not a segment.
*/
acces_gra *
graphicTablesManager::getBoxTableElem (graphObj_type obj)
{
  acces_gra *pt;

  if (obj < 0)
    {
      printf ("\n##INTERNAL ERROR## in graphicTablesManager::getBoxTableElem: illegal negative value %i for an object adress\n", obj);
      fflush (stdout);
      return null;
    }

  pt = tab_acces[obj % taille_tab];
  while (pt != null)
    {
      if (pt->graphObj == obj)
	return pt;;
      pt = pt->suivant;
    }
  return null;
}


/**
  Returns the data structure associated with the segment seg (in)
*/
acces_seg *
graphicTablesManager::getSegTableElem (int seg)
{
  acces_seg *pt;

  pt = t_acces_seg[seg % taille_tab];
  while (pt != null)
    {
      if (pt->id == seg)
	return pt;
      pt = pt->suivant;
    }
  return null;
}


/**
  Initialisation of the 3 hash tables
*/
void
graphicTablesManager::init_convnum (void)
{
  int i;

  for (i = 0; i < taille_t_hash; i++)
    {
      this_boxHashTable[i] = null;
      this_segmentHashTable[i] = null;
      this_pathHashTable[i] = null;
    }
}


/**
  Returns the conversion of num (in)
  tab (in) : 0 -> the search is done in the box hash table, 
           : 1 -> the search is done in the segment hash table, 
	   : 2 -> the search is done in the path hash table.
*/
int
graphicTablesManager::locinconvnum (int num, int tab)
{
  t_hash_num *listTail;
  int h, result;

  h = hashnum (num);
  switch (tab)
    {
    case 0:
      listTail = this_boxHashTable[h];
      break;
    case 1:
      listTail = this_segmentHashTable[h];
      break;
    case 2:
      listTail = this_pathHashTable[h];
      break;
    }
  if (listTail == null)
    return crtconvnum (tab, num, h, null);
  if (num < listTail->ovovovov)
    return crtconvnum (tab, num, h, listTail);
  return numinlconv (num, listTail);
}


/**
  Conversion of the box number and mofification in its tree p_design (in).
  Returns the new number of the box.
  This method is called when a new file is loaded and for a tree copy in
  order that objects numbers remain unique.
  */
int
graphicTablesManager::convnumbte (ntrees p_design)
{
  int n;

  n = locinconvnum ((int) nintof (p_design), 0);
  nmodifvalint (p_design, n);
  return (n);
}


/**
  Same method as convnumbte but for the ports.
*/
int
graphicTablesManager::convnumport (ntrees p_design)
{
  return (convnumbte (p_design));
}


/**
  Same method as convnumbte but for the links.
*/
int
graphicTablesManager::convnumlien (ntrees p_design)
{
  return (convnumbte (p_design));
}


/**
  Same method as convnumbte but for the segments.
*/
int
graphicTablesManager::convnumseg (ntrees p_design)
{
  int n;

  n = locinconvnum ((int) nintof (p_design), 1);
  nmodifvalint (p_design, n);
  return (n);
}


/**
  Same method as convnumbte but for the paths.
*/
int
graphicTablesManager::convnumchem (ntrees p_design)
{
  int n;

  n = locinconvnum ((int) nintof (p_design), 2);
  nmodifvalint (p_design, n);
  return (n);
}


/**
  Sets the halo of the acces_seg elem (in) with halo (in)
 */
void
graphicTablesManager::setSegElemWithHalo (acces_seg * elem, typcoord * halo)
{
  elem->ext1_halo.x = halo[0].x;
  elem->ext1_halo.y = halo[0].y;
  elem->ext2_halo.x = halo[1].x;
  elem->ext2_halo.y = halo[1].y;
}


/**
  Sets the halo of the acces_gra elem (in) with halo (in)
*/
void
graphicTablesManager::setBoxElemWithHalo (acces_gra * elem, typcoord * halo)
{
  elem->ext1_halo.x = halo[0].x;
  elem->ext1_halo.y = halo[0].y;
  elem->ext2_halo.x = halo[1].x;
  elem->ext2_halo.y = halo[1].y;
}


/**
  sets halo (out) with the halo contained in the acces_gra table elem (in)
 */
void
graphicTablesManager::setHaloWithBoxElem (acces_gra * elem, typcoord * halo)
{
  halo[0].x = elem->ext1_halo.x;
  halo[0].y = elem->ext1_halo.y;
  halo[1].x = elem->ext2_halo.x;
  halo[1].y = elem->ext2_halo.y;
}


/**
  set halo (out) with the halo contained in the acces_seg table elem (in)
 */
void
graphicTablesManager::setHaloWithSegElem (acces_seg * elem, typcoord * halo)
{
  halo[0].x = elem->ext1_halo.x;
  halo[0].y = elem->ext1_halo.y;
  halo[1].x = elem->ext2_halo.x;
  halo[1].y = elem->ext2_halo.y;
}


/**
  Inserts the graphical object objet (in) in the table acces_gra.
  The object may be a box, a port, a link or a path.
  graphTree (in) : the tree associated with the object
  halo (in) : its halo
  */
void
graphicTablesManager::inser_obj (graphObj_type objet, ntrees graphTree, typcoord * halo)
{
  int n, t;
  acces_gra *pt, *ptelem, *aux;

  n = objet % taille_tab;
  pt = tab_acces[n];
  t = sizeof (acces_gra);
  ptelem = (acces_gra *) calloc (1, t);
  ptelem->graphObj = objet;
  ptelem->elem_gra = graphTree;
  setBoxElemWithHalo (ptelem, halo);
  if (pt == null)
    tab_acces[n] = ptelem;
  else
    {
      aux = tab_acces[n];
      tab_acces[n] = ptelem;
      ptelem->suivant = aux;
    }
}


/**
  Removes the object objet (in) from the acces_gra table.
 */
void
graphicTablesManager::enlever_obj (graphObj_type objet)
{
  int n, trouve;
  acces_gra *pt, *pred;

  n = objet % taille_tab;
  pt = tab_acces[n];
  trouve = pk_false;
  while (!trouve && pt != null)
    if (pt->graphObj == objet)
      trouve = pk_true;
    else
      {
	pred = pt;
	pt = pt->suivant;
      }
  if (trouve)
    {
      if (pt == tab_acces[n])
	tab_acces[n] = pt->suivant;
      else
	pred->suivant = pt->suivant;
    }
}


/**
  Returns the tree associated with the graphical object obj (in). 
  The object may be a box, a port, a link or a path but not a segment; 
  for a segment, use the method acces_segment.
*/
ntrees
graphicTablesManager::graphObj_get_tree (graphObj_type obj)
{
  acces_gra *pt;

  pt = getBoxTableElem (obj);
  if (pt == null)
    return null;
  return (pt->elem_gra);
}


/**
  Inserts the segment seg (in) in the acces_seg table.
  direc_seg (in) : segment direction (vertic or horiz)
  pt_neg (in): number of the diffusor port
  pSeg (in) : tree associated with the segment
  halo (in) : its halo
  */
void
graphicTablesManager::inser_seg (int seg, int direc_seg, int pt_neg, ntrees pSeg, typcoord * halo)
{
  int n, t;
  acces_seg *pt, *ptelem, *aux;

  n = seg % taille_tab;
  pt = t_acces_seg[n];
  t = sizeof (acces_seg);
  ptelem = (acces_seg *) calloc (1, t);
  ptelem->id = seg;
  ptelem->arb_seg = pSeg;
  ptelem->port_ = pt_neg;
  ptelem->direc = direc_seg;
  setSegElemWithHalo (ptelem, halo);
  if (pt == null)
    t_acces_seg[n] = ptelem;
  else
    {
      aux = t_acces_seg[n];
      t_acces_seg[n] = ptelem;
      ptelem->suivant = aux;
    }
}


/**
  Removes the segment seg (in) from the acces_seg table.
*/
void
graphicTablesManager::enlever_seg (int seg)
{
  int n, trouve;
  acces_seg *pt, *pred;

  n = seg % taille_tab;
  pt = t_acces_seg[n];
  trouve = pk_false;
  while (!trouve && pt != null)
    if (pt->id == seg)
      trouve = pk_true;
    else
      {
	pred = pt;
	pt = pt->suivant;
      }
  if (trouve)
    {
      if (pt == t_acces_seg[n])
	t_acces_seg[n] = pt->suivant;
      else
	pred->suivant = pt->suivant;
    }
}


/**
  Return the tree associated with the segment seg (in)
*/
ntrees
graphicTablesManager::acces_segment (int seg)
{
  acces_seg *pt;

  pt = getSegTableElem (seg);
  if (pt != null)
    return (pt->arb_seg);
  else
    return (null);
}


/**
  Returns the halo of the segment seg (in)
*/
ty_halo *
graphicTablesManager::recup_segment_halo (int seg)
{
  acces_seg *pt;

  pt = getSegTableElem (seg);
  if (pt != null)
    {
      setHaloWithSegElem (pt, halo);
      return (&halo);
    }
  else
    return (null);
}


/**
 Returns the halo of the graphical object obj (in)
*/
ty_halo *
graphicTablesManager::recup_bte_halo (int obj)
{
  acces_gra *pt;

  pt = getBoxTableElem (obj);
  if (pt == null)
    return (null);
  setHaloWithBoxElem (pt, halo);
  return &halo;
}


/**
  Reset the halo of the segment seg (in) in the segments table.
*/
void
graphicTablesManager::annuler_halo_seg (int seg)
{
  acces_seg *pt;

  pt = getSegTableElem (seg);
  if (pt != null)
    {
      pt->ext1_halo.x = 0;
      pt->ext1_halo.y = 0;
      pt->ext2_halo.x = 0;
      pt->ext2_halo.y = 0;
    }
}


/**
  Set the halo of the segment seg (int) with the value halo (in).
  This method is called by the method "graphicalTrees::annuler_halo_chemins"
*/
void
graphicTablesManager::setSegHalo (int seg, typcoord * halo)
{
  acces_seg *pt;

  pt = getSegTableElem (seg);
  if (pt != null)
    if ((pt->ext1_halo.x == 0) || (pt->ext1_halo.y == 0) || (pt->ext2_halo.x == 0) || (pt->ext2_halo.y == 0))
      setSegElemWithHalo (pt, halo);
}
